(* give names to intermediate values (K-normalization) *)

type t = (* Kμ (caml2html: knormal_t) *)
  | Unit
  | Int of int
  | Float of float
  | Neg of Id.t
  | Add of Id.t * Id.t
  | Sub of Id.t * Id.t
  | FNeg of Id.t
  | FAdd of Id.t * Id.t
  | FSub of Id.t * Id.t
  | FMul of Id.t * Id.t
  | FDiv of Id.t * Id.t
  | IfEq of Id.t * Id.t * t * t (*  + ʬ (caml2html: knormal_branch) *)
  | IfLE of Id.t * Id.t * t * t (*  + ʬ *)
  | Let of (Id.t * Type.t) * t * t
  | Var of Id.t
  | LetRec of fundef * t
  | App of Id.t * Id.t list
  | Tuple of Id.t list
  | LetTuple of (Id.t * Type.t) list * Id.t * t
  | Get of Id.t * Id.t
  | Put of Id.t * Id.t * Id.t
  | ExtArray of Id.t
  | ExtFunApp of Id.t * Id.t list
and fundef = { name : Id.t * Type.t; args : (Id.t * Type.t) list; body : t }

let rec fv = function (* ˽иʼͳʡѿ (caml2html: knormal_fv) *)
  | Unit | Int(_) | Float(_) | ExtArray(_) -> S.empty
  | Add(x, y) | Sub(x, y) | FAdd(x, y) | FSub(x, y) | FMul(x, y) | FDiv(x, y) | Get(x, y) -> S.of_list [x; y]
  | Neg(x) | FNeg(x) -> S.singleton x
  | IfEq(x, y, e1, e2) | IfLE(x, y, e1, e2) -> S.add x (S.add y (S.union (fv e1) (fv e2)))
  | Let((x, t), e1, e2) -> S.union (fv e1) (S.remove x (fv e2))
  | Var(x) -> S.singleton x
  | LetRec({ name = (x, t); args = yts; body = e1 }, e2) ->
      let zs = S.diff (fv e1) (S.of_list (List.map fst yts)) in
      S.diff (S.union zs (fv e2)) (S.singleton x)
  | App(x, ys) -> S.of_list (x :: ys)
  | Tuple(xs) | ExtFunApp(_, xs) -> S.of_list xs
  | Put(x, y, z) -> S.of_list [x; y; z]
  | LetTuple(xs, y, e) -> S.add y (S.diff (fv e) (S.of_list (List.map fst xs)))

let insert_let (e, t) k = (* letؿ (caml2html: knormal_insert) *)
  match e with
  | Var(x) -> k x
  | _ ->
      let x = Id.gentmp t in
      let e', t' = k x in
      Let((x, t), e, e'), t'

let rec g env = function (* K롼 (caml2html: knormal_g) *)
  | Syntax.Unit -> Unit, Type.Unit
  | Syntax.Bool(b) -> Int(if b then 1 else 0), Type.Int (* true, false1, 0Ѵ (caml2html: knormal_bool) *)
  | Syntax.Int(i) -> Int(i), Type.Int
  | Syntax.Float(d) -> Float(d), Type.Float
  | Syntax.Not(e) -> g env (Syntax.If(e, Syntax.Bool(false), Syntax.Bool(true)))
  | Syntax.Neg(e) -> 
      insert_let (g env e) (fun x -> Neg(x), Type.Int)
  | Syntax.Add(e1, e2) -> (* ­K (caml2html: knormal_add) *)      
      insert_let (g env e1)
	(fun x -> insert_let (g env e2)
	    (fun y -> Add(x, y), Type.Int))
  | Syntax.Sub(e1, e2) ->
      insert_let (g env e1)
	(fun x -> insert_let (g env e2)
	    (fun y -> Sub(x, y), Type.Int))
  | Syntax.FNeg(e) ->
      insert_let (g env e)
	(fun x -> FNeg(x), Type.Float)
  | Syntax.FAdd(e1, e2) ->
      insert_let (g env e1)
	(fun x -> insert_let (g env e2)
	    (fun y -> FAdd(x, y), Type.Float))
  | Syntax.FSub(e1, e2) ->
      insert_let (g env e1)
	(fun x -> insert_let (g env e2)
	    (fun y -> FSub(x, y), Type.Float))
  | Syntax.FMul(e1, e2) ->
      insert_let (g env e1)
	(fun x -> insert_let (g env e2)
	    (fun y -> FMul(x, y), Type.Float))
  | Syntax.FDiv(e1, e2) ->
      insert_let (g env e1)
	(fun x -> insert_let (g env e2)
	    (fun y -> FDiv(x, y), Type.Float))
  | Syntax.Eq _ | Syntax.LE _ as cmp ->
      g env (Syntax.If(cmp, Syntax.Bool(true), Syntax.Bool(false)))
  | Syntax.If(Syntax.Not(e1), e2, e3) -> g env (Syntax.If(e1, e3, e2)) (* notˤʬѴ (caml2html: knormal_not) *)
  | Syntax.If(Syntax.Eq(e1, e2), e3, e4) ->
      insert_let (g env e1)
	(fun x -> insert_let (g env e2)
	    (fun y ->
	      let e3', t3 = g env e3 in
	      let e4', t4 = g env e4 in
	      IfEq(x, y, e3', e4'), t3))
  | Syntax.If(Syntax.LE(e1, e2), e3, e4) ->
      insert_let (g env e1)
	(fun x -> insert_let (g env e2)
	    (fun y ->
	      let e3', t3 = g env e3 in
	      let e4', t4 = g env e4 in
	      IfLE(x, y, e3', e4'), t3))
  | Syntax.If(e1, e2, e3) -> g env (Syntax.If(Syntax.Eq(e1, Syntax.Bool(false)), e3, e2)) (* ӤΤʤʬѴ (caml2html: knormal_if) *)
  | Syntax.Let((x, t), e1, e2) ->
      let e1', t1 = g env e1 in
      let e2', t2 = g (M.add x t env) e2 in
	Let((x, t), e1', e2'), t2
  | Syntax.Var(x) when M.mem x env -> Var(x), M.find x env
  | Syntax.Var(x) -> (* λ (caml2html: knormal_extarray) *)
      (match M.find x !Typing.extenv with
      | Type.Array(_) as t -> ExtArray x, t
      | _ -> failwith (Printf.sprintf "external variable %s does not have an array type" x))
  | Syntax.LetRec({ Syntax.name = (x, (_, t)); Syntax.args = yts; Syntax.body = e1 }, e2) ->
      let env' = M.add x t env in
      let e2', t2 = g env' e2 in
      let e1', t1 = g (M.add_list yts env') e1 in
      LetRec({ name = (x, t); args = yts; body = e1' }, e2'), t2
  | Syntax.App(Syntax.Var(f), e2s) when not (M.mem f env) -> (* ؿθƤӽФ (caml2html: knormal_extfunapp) *)
      (match M.find f !Typing.extenv with
      | Type.Fun(_, _, t, _) ->
	  let rec bind xs = function (* "xs" are identifiers for the arguments *)
	    | [] -> ExtFunApp(f, xs), t
	    | e2 :: e2s ->
		insert_let (g env e2)
		  (fun x -> bind (xs @ [x]) e2s) in
	  bind [] e2s (* left-to-right evaluation *)
      | _ -> assert false)
  | Syntax.App(e1, e2s) ->
      (match g env e1 with
      | _, Type.Fun(_, _, t, _) as g_e1 ->
	  insert_let g_e1
	    (fun f ->
	      let rec bind xs = function (* "xs" are identifiers for the arguments *)
		| [] -> App(f, xs), t
		| e2 :: e2s ->
		    insert_let (g env e2)
		      (fun x -> bind (xs @ [x]) e2s) in
	      bind [] e2s) (* left-to-right evaluation *)
      | _, t -> 
	  (Syntax.print (Syntax.App (e1, e2s)); 
	   print_newline (); print_string (Type.t_to_string t);
	   print_newline (); assert false))
  | Syntax.Tuple(es) ->
      let rec bind xs ts = function (* "xs" and "ts" are identifiers and types for the elements *)
	| [] -> Tuple(xs), Type.Tuple(ts)
	| e :: es ->
	    let _, t as g_e = g env e in
	    insert_let g_e
	      (fun x -> bind (xs @ [x]) (ts @ [t]) es) in
      bind [] [] es
  | Syntax.LetTuple(xts, e1, e2) ->
      insert_let (g env e1)
	(fun y ->
	  let e2', t2 = g (M.add_list xts env) e2 in
	  LetTuple(xts, y, e2'), t2)
  | Syntax.Array(e1, e2) ->
      insert_let (g env e1)
	(fun x ->
	  let _, t2 as g_e2 = g env e2 in
	  insert_let g_e2
	    (fun y ->
	      let l =
		match t2 with
		| Type.Float -> "create_float_array"
		| _ -> "create_array" in
	      ExtFunApp(l, [x; y]), Type.Array(t2)))
  | Syntax.Get(e1, e2) ->
      (match g env e1 with
      |	_, Type.Array(t) as g_e1 ->
	  insert_let g_e1
	    (fun x -> insert_let (g env e2)
		(fun y -> Get(x, y), t))
      | _ -> assert false)
  | Syntax.Put(e1, e2, e3) ->
      insert_let (g env e1)
	(fun x -> insert_let (g env e2)
	    (fun y -> insert_let (g env e3)
		(fun z -> Put(x, y, z), Type.Unit)))
  | Syntax.Empty -> Int(0), Type.Int
  | Syntax.Cons(e1, e2) -> 
      let _, t1 as g_e1 = g env e1 in
      let _, t2 as g_e2 = g env e2 in
	insert_let g_e1 
	  (fun x -> insert_let g_e2 
	     (fun y -> Tuple([x; y]), Type.Tuple([t1; t2])))
  | Syntax.Match(lst, e1, (((y, t) as yt), ((z, s) as zs), e2)) ->
      insert_let (g env lst) 
	(fun x -> 
	   let e1', t1 = g env e1 in
	   let e2', t2 = g (M.add y t (M.add z s env)) e2 in
	     insert_let (Int(0), Type.Int) 
	       (fun y -> IfEq(y, x, e1', (LetTuple([yt; zs], x, e2'))), t2))
  (* external function Ѵ *)
  | Syntax.Shift((x, (_, t)), e, t') -> 
      let e, t'' = g (M.add x t env) e in
      let v = Id.genid "s" in
      let t1 = 
	match t with Type.Fun ([t1], _, _, _) -> t1 | _ -> assert false in
      let tv = Type.Fun ([t], t'', t'',t') in
	LetRec({ name = (v, tv); args = [(x, t)]; body = e }, 
	       ExtFunApp("shift", [v])), t1
  | Syntax.Reset(e, t) -> 
      let e, t' = g env e in
      let v = Id.genid "r" and x = Id.genid "x" in
      let fdef = { name = (v, Type.Fun([Type.Unit], t, t', t)); 
		   args = [(x, Type.Unit)]; body = e } in
	LetRec(fdef, ExtFunApp("reset", [v])), t

let f e = fst (g M.empty e)


(* print *)
let xt2str (x, t) = "(" ^ x ^ " : " ^ (Type.t_to_string t) ^ ")" (* x *)
let xts2str (x, ts) = "(" ^ x ^ " : " ^ (Type.ts_to_string ts) ^ ")" (* x *)

let rec print = function 
  | Unit -> print_string "()"
  | Int i -> print_int i
  | Float f -> print_float f
  | Neg x | FNeg x -> print_string ("-" ^ x)
  | Add (x, y) -> print_string (x ^ " + " ^ y)
  | Sub (x, y) -> print_string (x ^ " - " ^ y)
  | FAdd (x, y) -> print_string (x ^ " +. " ^ y)
  | FSub (x, y) -> print_string (x ^ " -. " ^ y)
  | FMul (x, y) -> print_string (x ^ " *. " ^ y)
  | FDiv (x, y) -> print_string (x ^ " /. " ^ y)
  | IfEq (x, y, e1, e2) -> 
      print_string ("if " ^ x ^ " = " ^ y ^ "\n  then ");
      print e1; print_string "\n  else "; print e2
  | IfLE (x, y, e1, e2) ->
      print_string ("if " ^ x ^ " <= " ^ y ^ "\n  then ");
      print e1; print_string "\n  else "; print e2
  | Let (xt, e1, e2) ->
      print_string ("let " ^ xt2str xt ^ " = "); print e1;
      print_string " in\n  "; print e2
  | Var x -> print_string x
  | LetRec ({ name = xt; args = []; body = e1 }, e2) ->
      print_string ("let rec " ^ (xt2str xt) ^ " = "); print e1; 
      print_string " in\n  "; print e2
  | LetRec ({ name = xt; args = yts; body = e1 }, e2) ->
      print_string ("let rec " ^ (xt2str xt));
      List.iter (fun yt -> print_string (" " ^ xt2str yt)) yts;
      print_string " = "; print e1; print_string " in\n  "; print e2
  | App (x, ys) | ExtFunApp (x, ys) ->
      print_string x; 
      List.iter (fun y -> print_string (" " ^ y)) ys
  | Tuple [] | LetTuple ([], _, _) -> assert false
  | Tuple (x :: rest) ->
      print_string ("(" ^ x);
      List.iter (fun x -> print_string (", " ^ x)) rest; print_string ")"
  | LetTuple (((x, _) :: xts), y, e) -> 
      print_string ("let (" ^ x);
      List.iter (fun (x, _) -> print_string (", " ^ x)) xts; 
      print_string (") = " ^ y ^ " in\n  "); print e      
  | Get (x, y) -> Printf.printf "%s.(%s)" x y
  | Put (x, y, z) -> Printf.printf "%s.(%s) <- %s" x y z
  | ExtArray x -> print_string x
