type t = (* MinCamlιʸɽǡ (caml2html: syntax_t) *)
  | Unit
  | Bool of bool
  | Int of int
  | Float of float
  | Not of t
  | Neg of t
  | Add of t * t
  | Sub of t * t
  | FNeg of t
  | FAdd of t * t
  | FSub of t * t
  | FMul of t * t
  | FDiv of t * t
  | Eq of t * t
  | LE of t * t
  | If of t * t * t
  | Let of (Id.t * Type.t) * t * t
  | Var of Id.t
  | LetRec of fundef * t
  | App of t * t list
  | Tuple of t list
  | LetTuple of (Id.t * Type.t) list * t * t
  | Array of t * t
  | Get of t * t
  | Put of t * t * t
  | Empty
  | Cons of t * t
  | Match of t * t * ((Id.t * Type.t) * (Id.t * Type.t) * t)
  (* K ѴΤθη⤯ *)
  | Shift of (Id.t * Type.ts) * t * Type.t 
  | Reset of t * Type.t   (* K ѴΤ reset e η⤯ *)
and fundef = { name : Id.t * Type.ts; args : (Id.t * Type.t) list; body : t }

let xt2str (x, t) = "(" ^ x ^ " : " ^ (Type.t_to_string t) ^ ")"
let xts2str (x, ts) = "(" ^ x ^ " : " ^ (Type.ts_to_string ts) ^ ")"

let rec to_string = function
  | Unit -> "()"
  | Bool b -> string_of_bool b
  | Int i -> string_of_int i
  | Float f -> string_of_float f
  | Not e -> "not " ^ to_string e
  | Neg e | FNeg e -> "-" ^ to_string e
  | Add (e1, e2) -> "(" ^ (to_string e1) ^ " + " ^ (to_string e2) ^ ")"
  | Sub (e1, e2) -> "(" ^ (to_string e1) ^ " - " ^ (to_string e2) ^ ")"
  | FAdd (e1, e2) -> "(" ^ (to_string e1) ^ " +. " ^ (to_string e2) ^ ")"
  | FSub (e1, e2) -> "(" ^ (to_string e1) ^ " -. " ^ (to_string e2) ^ ")"
  | FMul (e1, e2) -> "(" ^ (to_string e1) ^ " *. " ^ (to_string e2) ^ ")"
  | FDiv (e1, e2) -> "(" ^ (to_string e1) ^ " /. " ^ (to_string e2) ^ ")"
  | Eq (e1, e2) -> (to_string e1) ^ " = " ^ (to_string e2)
  | LE (e1, e2) -> (to_string e1) ^ " <= " ^ (to_string e2)
  | If (e1, e2, e3) -> 
      "if " ^ (to_string e1) ^ 
	"\n  then " ^ (to_string e2) ^ "\n  else " ^ (to_string e3)
  | Let (xt, e1, e2)  -> 
      "let " ^ (xt2str xt) ^ " = " ^ 
	(to_string e1) ^ " in \n  " ^ (to_string e2)
  | Var (x) -> x
  | LetRec ({ name = _ ; args = []; body = _ }, _) -> failwith "empty argument"
  | LetRec ({ name = xts; args = at :: rest; body = e1 }, e2) ->
      "let rec " ^ (xts2str xts) ^ " " ^
	(List.fold_left (fun s xt -> s ^ " " ^ (xt2str xt)) (xt2str at) rest) ^
	" = \n  " ^ (to_string e1) ^ " in\n\n  " ^ (to_string e2)
  | App (e, es) -> 
      (to_string e) ^ " " ^ 
	(List.fold_left (fun s e -> s ^ (to_string e) ^ " ") "" es)
  | Tuple ([]) -> failwith "empty tuple"
  | Tuple (e :: es) -> 
      List.fold_left (fun s e -> s ^ " * " ^ (to_string e)) (to_string e) es
  | LetTuple ([], e1, e2) -> failwith "empty tuple"
  | LetTuple (xt :: xts, e1, e2) -> 
      let s = 
	List.fold_left (fun s xt -> s ^ ", " ^ (xt2str xt)) (xt2str xt) xts in
	"let " ^ s ^ " = " ^ (to_string e1) ^ " in\n  " ^ (to_string e2)
  | Array (e1, e2) ->
      "Array.create " ^ (to_string e1) ^ " " ^ (to_string e2)
  | Get (e1, e2) -> 
      (to_string e1) ^ ".(" ^ (to_string e2) ^ ")"
  | Put (e1, e2, e3) ->
      (to_string e1) ^ ".(" ^ (to_string e2) ^ ") <- " ^ (to_string e3)
  | Empty -> "[]"
  | Cons (e1, e2) -> (to_string e1) ^ " :: " ^ (to_string e2)
  | Match (e1, e2, (xt, yt, e3)) -> 
      "match " ^ (to_string e1) ^ " with \n    [] -> " ^ (to_string e2) ^
	"\n  | " ^ (xt2str xt) ^ " :: " ^ (xt2str yt) ^ " -> " ^ (to_string e3)
  | Shift (xts, e, _) ->
      "shift (fun " ^ (xts2str xts) ^ " -> " ^ (to_string e) ^ ")"
  | Reset (e, _) ->
      "<" ^ (to_string e) ^ ">"

let rec print = function
  | Unit -> print_string "()"
  | Bool b -> print_string (string_of_bool b)
  | Int i -> print_int i
  | Float f -> print_float f
  | Not e -> print_string "not " ;print e
  | Neg e | FNeg e -> print_string "-" ; print e
  | Add (e1, e2) -> 
      print_string "("; print e1; print_string " + "; 
      print e2; print_string ")"
  | Sub (e1, e2) -> 
      print_string "("; print e1; print_string " - "; 
      print e2; print_string ")"
  | FAdd (e1, e2) -> 
      print_string "("; print e1; print_string " +. "; 
      print e2; print_string ")"
  | FSub (e1, e2) ->
      print_string "("; print e1; print_string " -. "; 
      print e2; print_string ")"
  | FMul (e1, e2) ->
      print_string "("; print e1; print_string " *. "; 
      print e2; print_string ")"
  | FDiv (e1, e2) ->
      print_string "("; print e1; print_string " /. "; 
      print e2; print_string ")"
  | Eq (e1, e2) ->
      print_string "("; print e1; print_string " = "; 
      print e2; print_string ")"
  | LE (e1, e2) ->
      print_string "("; print e1; print_string " <= "; 
      print e2; print_string ")"
  | If (e1, e2, e3) ->
      print_string "if "; print e1; print_string "\nthen ";
      print e2; print_string "\nelse"; print e3
  | Let (xt, e1, e2) ->
      print_string ("let " ^ xt2str xt ^ " = "); print e1;
      print_string " in \n"; print e2
  | Var x -> print_string x
  | LetRec ({ name = xts; args = args; body = e1 }, e2) ->
      print_string ("let rec " ^ xts2str xts ^ " \n\t");
      List.iter (fun xt -> print_string (" " ^ xt2str xt)) args;
      print_string " =\n  ";
      print e1;
      print_string " in \n";
      print e2
  | App (f, args) ->
      print_string "("; print f; 
      List.iter (fun x -> print_string " "; print x) args; print_string ")"
  | Tuple [] | LetTuple ([], _, _) -> failwith "empty tuple"
  | Tuple (a :: rest) ->
      print_string "("; print a;
      List.iter (fun x -> print_string ", "; print x) rest; print_string ")"
  | LetTuple (((a, _) :: rest), e1, e2) ->
      print_string "let ("; print_string a;
      List.iter (fun (x, _) -> print_string (", " ^ x)) rest; 
      print_string " = "; print e1; print_string " in\n  "; print e2
  | Array (e1, e2) -> 
      print_string "Array.make "; print e1; print_string " "; print e2
  | Get (e1, e2) ->
      print e1; print_string ".("; print e2; print_string ")"
  | Put (e1, e2, e3) ->
      print (Get (e1, e2)); print_string " <- "; print e3
  | Empty -> print_string "[]"
  | (Cons (_, _)) as lst -> 
      (let rec loop = function
	 | Empty -> ()
	 | Cons (e1, Empty) -> print e1
	 | Cons (e1, e2) -> print e1; print_string "; "; loop e2
	 | e -> print e in
	 print_string "["; loop lst; print_string "]") 
  | Match (lst, e1, (a, rest, e2)) ->
      print_string "match "; print lst; print_string " with \n    [] -> "; 
      print e1; print_string "\n  | "; print_string (xt2str a);
      print_string " :: "; print_string (xt2str rest);
      print_string " -> "; print e2
  | Shift (xts, e, _) -> 
      print_string ("shift (fun " ^ (xts2str xts) ^ " -> ");
      print e; print_string ")"
  | Reset (e, _) -> print_string "reset ("; print e; print_string ")"

      
