(* type inference/reconstruction *)

open Syntax

exception Unify of Type.t * Type.t
exception Error of t * Type.t * Type.t

let extenv = ref M.empty

(* for pretty printing (and type normalization) *)
let rec deref_typ = function (* ѿȤǤؿ (caml2html: typing_deref) *)
  | Type.Fun(t1s, t2, t3, t4) ->
      Type.Fun(List.map deref_typ t1s, deref_typ t2, deref_typ t3, deref_typ t4)
  | Type.Tuple(ts) -> Type.Tuple(List.map deref_typ ts)
  | Type.Array(t) -> Type.Array(deref_typ t)
  | Type.Var({ contents = Some(t) } as r) ->
      let t' = deref_typ t in
      r := Some(t');
      t'
  | Type.List(t) -> Type.List(deref_typ t)
  | t -> t
let deref_ts (rs_ref, t) = (rs_ref, deref_typ t)
let rec deref_id_typ (x, t) = (x, deref_typ t)
let rec deref_id_ts (x, ts) = (x, deref_ts ts)
let rec deref_term = function
  | Not(e) -> Not(deref_term e)
  | Neg(e) -> Neg(deref_term e)
  | Add(e1, e2) -> Add(deref_term e1, deref_term e2)
  | Sub(e1, e2) -> Sub(deref_term e1, deref_term e2)
  | Eq(e1, e2) -> Eq(deref_term e1, deref_term e2)
  | LE(e1, e2) -> LE(deref_term e1, deref_term e2)
  | FNeg(e) -> FNeg(deref_term e)
  | FAdd(e1, e2) -> FAdd(deref_term e1, deref_term e2)
  | FSub(e1, e2) -> FSub(deref_term e1, deref_term e2)
  | FMul(e1, e2) -> FMul(deref_term e1, deref_term e2)
  | FDiv(e1, e2) -> FDiv(deref_term e1, deref_term e2)
  | If(e1, e2, e3) -> If(deref_term e1, deref_term e2, deref_term e3)
  | Let(xt, e1, e2) -> Let(deref_id_typ xt, deref_term e1, deref_term e2)
  | LetRec({ name = xt; args = yts; body = e1 }, e2) ->
      LetRec({ name = deref_id_ts xt;
	       args = List.map deref_id_typ yts;
	       body = deref_term e1 },
	     deref_term e2)
  | App(e, es) -> App(deref_term e, List.map deref_term es)
  | Tuple(es) -> Tuple(List.map deref_term es)
  | LetTuple(xts, e1, e2) -> LetTuple(List.map deref_id_typ xts, deref_term e1, deref_term e2)
  | Array(e1, e2) -> Array(deref_term e1, deref_term e2)
  | Get(e1, e2) -> Get(deref_term e1, deref_term e2)
  | Put(e1, e2, e3) -> Put(deref_term e1, deref_term e2, deref_term e3)
  | Cons(e1, e2) -> Cons(deref_term e1, deref_term e2)
  | Match(e1, e2, (xs, ys, e3)) ->
      Match(deref_term e1, deref_term e2, 
	    (deref_id_typ xs, deref_id_typ ys, deref_term e3))
  | Shift(xts, e, t) -> Shift(deref_id_ts xts, deref_term e, deref_typ t)
  | Reset(e, t) -> Reset(deref_term e, deref_typ t)
  | e -> e

let rec occur r1 = function (* occur check (caml2html: typing_occur) *)
  | Type.Fun(t2s, t2, t3, t4) -> 
      List.exists (occur r1) t2s || occur r1 t2 || occur r1 t3 || occur r1 t4
  | Type.Tuple(t2s) -> List.exists (occur r1) t2s
  | Type.Array(t2) -> occur r1 t2
  | Type.Var(r2) when r1 == r2 -> true
  | Type.Var({ contents = None }) -> false
  | Type.Var({ contents = Some(t2) }) -> occur r1 t2
  | Type.List(t2) -> occur r1 t2
  | _ -> false

(* μͳѿ *)
(* free_typ : Type.t -> Type.t option ref Type.S.t *)
let rec free_typ = function 
  | Type.Fun(t1s, t2, t3, t4) ->
      List.fold_left (fun s t -> Type.S.union (free_typ t) s) 
	Type.S.empty (t2 :: t3 :: t4 :: t1s)
  | Type.Tuple(ts) -> 
      List.fold_left (fun s t -> Type.S.union (free_typ t) s) Type.S.empty ts
  | Type.Array(t) | Type.List(t) -> free_typ t
  | Type.Var({ contents = None } as r) -> Type.S.singleton r
  | Type.Var({ contents = Some(t) }) -> free_typ t
  | _ -> Type.S.empty

(*  scheme μͳѿ *)
(* free_typ_s : Type.ts -> Type.t option ref Type.S.t *)
let free_typ_s (refs, t) =  Type.S.diff (free_typ t) !refs

(* Ķμͳѿ *)
(* free_typ_e : M.t -> Type.t option ref Type.S.t *)
let free_typ_e env = 
  M.fold (fun _ ts s -> Type.S.union (free_typ_s ts) s) env Type.S.empty

(* Gen(; )  *)
(* gen : Type.t -> M.t -> Type.ts *)
let gen t env =
  (ref (Type.S.diff (free_typ t) (free_typ_e env)), t)

(* refs Ѳ *)
(* close : Type.ts -> M.t -> unit *)
let close (refs, t) env = 
  refs := Type.S.diff (free_typ t) (free_typ_e env)

(*  scheme   ηѿ˿ѿ֤ *)
let instance (refs, t) = 
  Type.S.fold (fun r t -> Type.subst r (Type.gentyp ()) t) !refs t

let rec unify t1 t2 = (* 礦褦ˡѿؤ򤹤 (caml2html: typing_unify) *)
  match t1, t2 with
  | Type.Unit, Type.Unit | Type.Bool, Type.Bool | Type.Int, Type.Int | Type.Float, Type.Float -> ()
  | Type.Fun(t1s, t1', t1'', t1'''), Type.Fun(t2s, t2', t2'', t2''') ->
      (try List.iter2 unify t1s t2s
      with Invalid_argument("List.iter2") -> raise (Unify(t1, t2)));
      unify t1' t2';
      unify t1'' t2'';
      unify t1''' t2'''
  | Type.Tuple(t1s), Type.Tuple(t2s) ->
      (try List.iter2 unify t1s t2s
      with Invalid_argument("List.iter2") -> raise (Unify(t1, t2)))
  | Type.Array(t1), Type.Array(t2) | Type.List(t1), Type.List(t2) -> unify t1 t2
  | Type.Var(r1), Type.Var(r2) when r1 == r2 -> ()
  | Type.Var({ contents = Some(t1') }), _ -> unify t1' t2
  | _, Type.Var({ contents = Some(t2') }) -> unify t1 t2'
  | Type.Var({ contents = None } as r1), _ -> (* ̤ηѿξ (caml2html: typing_undef) *) 
      if occur r1 t2 then raise (Unify(t1, t2));
      r1 := Some(t2)
  | _, Type.Var({ contents = None } as r2) -> 
      if occur r2 t1 then raise (Unify(t1, t2));
      r2 := Some(t1)
  | _, _ -> raise (Unify(t1, t2))

let rec g env e = (* 롼 (caml2html: typing_g) *)
  (* returns (Type.t * Type.t * Type.t)  η()ηη() *)
  try
    match e with
    | Unit -> let t = Type.gentyp () in (t, Type.Unit, t)
      (* env; t |- () : Unit; t *)
    | Bool(_) -> let t = Type.gentyp () in (t, Type.Bool, t)
      (* env; t |- bool : Bool; t *)
    | Int(_) -> let t = Type.gentyp () in (t, Type.Int, t)
      (* env; t |- int : Int; t *)
    | Float(_) -> let t = Type.gentyp () in (t, Type.Float, t)
      (* env; t |- float : Float; t *)
    | Not(e) ->
      (* env; t1 |- not : [Bool]/t2 -> Bool/t2; t1
         env; t2 |- e : Bool; t1
         ---------------------------------------
         env; t2 |- not e : Bool; t1             *)
	let (t2, t, t1) = g env e in
	unify Type.Bool t;
	(t2, Type.Bool, t1)
    | Neg(e) ->
      (* env; t1 |- neg : [Int]/t2 -> Int/t2; t1
         env; t2 |- e : Int; t1
         -------------------------------------
         env; t2 |- neg e : Int; t1            *)
	let (t2, t, t1) = g env e in
	unify Type.Int t;
	(t2, Type.Int, t1)
    | Add(e1, e2) | Sub(e1, e2) -> (* ­ʤȰˤη (caml2html: typing_add) *)
      (* env; t1 |- add : [Int, Int]/t3 -> Int/t3; t1
         env; t2 |- e1 : Int; t1
	 env; t3 |- e2 : Int; t2
         --------------------------------------------
         env; t3 |- add e1 e2 : Int; t1               *)
	let (t2, i1, t1) = g env e1 in
	unify Type.Int i1;
	let (t3, i2, t2') = g env e2 in
	unify Type.Int i2;
        unify t2 t2';
	(t3, Type.Int, t1)
    | FNeg(e) ->
      (* env; t1 |- neg : [Float]/t2 -> Float/t2; t1
	 env; t2 |- e : Float; t1
	 -------------------------------------------
	 env; t2 |- neg e : Float; t1                *)
	let (t2, t, t1) = g env e in
	unify Type.Float t;
        (t2, Type.Float, t1)
    | FAdd(e1, e2) | FSub(e1, e2) | FMul(e1, e2) | FDiv(e1, e2) ->
      (* env; t1 |- fadd : [Float, Float]/t3 -> Float/t3; t1
	 env; t2 |- e1 : Float; t1
	 env; t3 |- e2 : Float; t2
	 ---------------------------------------------------
	 env; t3 |- fadd e1 e2 : Float; t1                   *)
	let (t2, f1, t1) = g env e1 in
	unify Type.Float f1;
        let (t3, f2, t2') = g env e2 in
        unify Type.Float f2;
        unify t2 t2';
        (t3, Type.Float, t1)
    | Eq(e1, e2) | LE(e1, e2) ->
      (* env; t1 |- equal : ['a, 'a]/t3 -> Bool/t3; t1
	 env; t2 |- e1 : 'a; t1
	 env; t3 |- e2 : 'a; t2
	 ---------------------------------------------
	 env; t3 |- equal e1 e2 : Bool; t1             *)
	let (t2, a, t1) = g env e1 in
        let (t3, a', t2') = g env e2 in
        unify a a';
        unify t2 t2';
        (t3, Type.Bool, t1)
    | If(e1, e2, e3) ->
      (* env; t1 |- e1 : Bool; t2
	 env; t3 |- e2 : 'a; t1
	 env; t3 |- e3 : 'a; t1
	 -----------------------------------------
	 env; t3 |- if e1 then e2 else e3 : 'a; t2 *)
	let (t1, b, t2) = g env e1 in
        unify Type.Bool b;
        let (t3, a, t1') = g env e2 in 
        unify t1 t1';
	let (t3', a', t1'') = g env e3 in 
        unify t1 t1'';
        unify t3 t3';
        unify a a';
        (t3, a, t2)
(*    | Let((x, t), e1, e2) -> (* letη (caml2html: typing_let) *)
	let (t1, t2, t3) = g env e1 in
	unify t t2;
	let (t4, t5, t1') = g (M.add x (Type.to_ts t2) env) e2 in
	unify t1 t1';
	(t4, t5, t3) *)
    | Let((x, t), e1, e2) -> (* let η (caml2html: typing_let) *)
      (* env;                      t2 |- e1 : t4; t2    (* e1 : pure *)
	 env, [x : Gen(t4; env)] ; t1 |- e2 : t3; t2
	 -------------------------------------------
	 env; t1 |- let x = e1 in e2 : t3; t2        *)
	let (t2, t4, t2') = g env e1 in
        unify t t4;
	unify t2 t2';
(*	let t = gen t4 env in  (* x η *) *)
	let t = Type.to_ts t4 in
	let (t1, t3, t2'') = g (M.add x t env) e2 in
	unify t2 t2'';
        (t1, t3, t2)
    | Var(x) when M.mem x env -> (* ѿη (caml2html: typing_var) *)
      (* env; t |- x : instance(env x); t *)
	let t = Type.gentyp () in (t, instance (M.find x env), t)
    | Var(x) when M.mem x !extenv -> 
      (* env; t |- x : (extenv x); t *)
	let t = Type.gentyp () in (t, M.find x !extenv, t)
    | Var(x) -> (* ѿη (caml2html: typing_extvar) *)
      (* ѿ monomolphic *)
      (* env; t |- x : new_t; t *)
	Format.eprintf "free variable %s assumed as external@." x;
	let t = Type.gentyp () in
	let t' = Type.gentyp () in
	extenv := M.add x t !extenv;
	(t', t, t')
    | LetRec({ name = (x, ts); args = yts; body = e1 }, e2) -> 
      (* let rec η (caml2html: typing_letrec) *)
      (* env[xt, yts]; t1 |- e1 : t2; t3
	 ts = close (yts/t1 -> t2/t3) env
	 env[x : ts] ; t4 |- e2 : t5; t6
	 -------------------------------------------
	 env; t4 |- let rec x ys = e1 in e2 : t5; t6 *)
	let env' = M.add x ts env in
	let env'' =
	  M.add_list (List.map (fun (y, t) -> (y, Type.to_ts t)) yts) env' in
	let (t1, t2, t3) = g env'' e1 in
	let (_, t) = ts in
        unify t (Type.Fun(List.map snd yts, t1, t2, t3));
        close ts env;
	let (t4, t5, t6) = g env' e2 in
        (t4, t5, t6)
    | App(e, es) -> (* ؿŬѤη (caml2html: typing_app) *)
      (* env; t3 |- e : (ts/t1 -> t/t2); t4
	 env; t2 |- es : ts; t3
	 -----------------------------------
	 env; t1 |- e es : t; t4             *)
	let t = Type.gentyp () in
	let t1 = Type.gentyp () in
	let (t2, ts, t3) = g_list env es in
        let (t3', t', t4) = g env e in
        unify t3 t3';
        unify t' (Type.Fun(ts, t1, t, t2));
        (t1, t, t4)
    | Tuple(es) -> 
      (* env; t1 |- es : ts; t2 *)
	let (t1, ts, t2) = g_list env es in
	(t1, Type.Tuple(ts), t2)
    | LetTuple(xts, e1, e2) ->
      (* env;             t1 |- e1 : Tuple(lst); t2
	 env, [xs : lst]; t3 |- e2 : t4; t1
	 ------------------------------------------
	 env; t3 |- let xs = e1 in e2 : t4; t2      *)
	let (t1, tlst, t2) = g env e1 in
        unify (Type.Tuple(List.map snd xts)) tlst;
        let env' =
	  M.add_list (List.map (fun (x, t) -> (x, Type.to_ts t)) xts) env in
	let (t3, t4, t1') = g env' e2 in
        unify t1 t1';
        (t3, t4, t2)
    | Array(e1, e2) -> (* must be a primitive for "polymorphic" typing *)
      (* env; t1 |- e1 : Int; t2
	 env; t3 |- e2 : 'a ; t1
	 ---------------------------------------------
	 env; t3 |- Array.create e1 e2 : Array('a); t2 *)
	let (t1, i, t2) = g env e1 in
        unify Type.Int i;
        let (t3, a, t1') = g env e2 in
        unify t1 t1';
        (t3, Type.Array(a), t2)
    | Get(e1, e2) ->
      (* env; t1 |- e1 : Array('a); t2
	 env; t3 |- e2 : Int; t1
	 -----------------------------
	 env; t3 |- e1.(e2) : 'a; t2   *)
	let (t1, array, t2) = g env e1 in
	let a = Type.gentyp () in
        unify (Type.Array(a)) array;
        let (t3, i, t1') = g env e2 in
        unify Type.Int i;
        unify t1 t1';
        (t3, a, t2)
    | Put(e1, e2, e3) ->
      (* env; t2 |- e1 : Array('a); t1
	 env; t3 |- e2 : Int; t2
	 env; t4 |- e3 : 'a; t3
	 -----------------------------------
	 env; t4 |- e1.(e2) <- e3 : Unit; t1 *)
	let (t2, array, t1) = g env e1 in
	let a = Type.gentyp () in
        unify (Type.Array(a)) array;
	let (t3, i, t2') = g env e2 in
        unify Type.Int i;
        unify t2 t2';
        let (t4, a', t3') = g env e3 in
        unify t3 t3';
        unify a a';
        (t4, Type.Unit, t3)
    | Empty -> 
      (* env; t |- [] : 'a list; t *)
	let a = Type.gentyp () in
	let t = Type.gentyp () in
        (t, Type.List(a), t)
    | Cons(e1, e2) ->
      (* env; t2 |- e1 : 'a; t1
	 env; t3 |- e2 : 'a list; t2
	 ---------------------------------
	 env; t3 |- e1 :: e2 : 'a list; t1 *)
	let (t2, a, t1) = g env e1 in
	let (t3, alst, t2') = g env e2 in
        unify t2 t2';
        unify (Type.List(a)) alst;
        (t3, alst, t1)
    | Match(e1, e2, ((x, t), (y, s), e3)) ->
      (* env;           t2 |- e1 : t list; t1
	 env;           t3 |- e2 : 'a; t2
	 env, [xt; ys]; t3 |- e3 : 'a; t2
	 -----------------------------------------------------------
	 env; t3 |- match e1 with [] -> e2 | xt :: ys -> e3 : 'a; t1 *)
	let (t2, tlst, t1) = g env e1 in
        unify (Type.List(t)) tlst;
	unify s tlst;
	let (t3, a, t2') = g env e2 in
        unify t2 t2';
	let (t3', a', t2'') = 
	  g (M.add x (Type.to_ts t) (M.add y (Type.to_ts s) env)) e3 in
        unify t3 t3';
        unify a a';
        unify t2 t2'';
        (t3, a, t1)
    | Shift((x, ts), e, t') ->
      (* env, [x:for_all t3.(t2/t3 -> t1/t3)]; t4 |- e : t4; t5
	 ------------------------------------------------------
         env; t1 |- shift (fun x -> e) : t2; t5                 *)
	let t1 = Type.gentyp () in
	let t2 = Type.gentyp () in
	let r3 = ref None in
	let t3 = Type.Var r3 in
	(* x η = ts ɡΤ褦ʷΤɬפ뤫 
	 * unify ơ S ΤۤѤƤɬפ          *)
	let (refs, t) = ts in
        unify t (Type.Fun([t2], t3, t1, t3));
        refs := Type.S.singleton r3;
	let (t4, t4', t5) = g (M.add x ts env) e in
        unify t4 t4';
	unify t' t5;
        (t1, t2, t5)
    | Reset(e, t) ->
      (* env; t1 |- e : t1; t2
         ---------------------------
         env; t3 |- reset e : t2; t3 *)
	let (t1, t1', t2) = g env e in
	unify t1 t1';
	let t3 = Type.gentyp () in
	unify t t2;
	(t3, t2, t3)
  with Unify(t1, t2) -> 
    let e = deref_term e and t1 = deref_typ t1 and t2 = deref_typ t2 in 
      print_string "at ";
      Syntax.print e; print_newline ();
      Printf.printf "%s and %s\n" (Type.t_to_string t1) (Type.t_to_string t2);
      raise (Error(e, t1, t2))

(* e ΥꥹȤɾ g *)
(* g_list : M.t -> Syntax.t -> (Type.t * Type.t list * Type.t) *)
and g_list env = function
  | [] -> let t = Type.gentyp () in (t, [], t)
    (* env; t |- [] : []; t *)
  | e :: es ->
    (* env; t1 |- e : 'a; t2
       env; t3 |- es : rest; t1
       -----------------------------------
       env; t3 |- e :: es : 'a :: rest; t2 *)
      let (t1, a, t2) = g env e in
      let (t3, rest, t1') = g_list env es in
      unify t1 t1';
      (t3, a :: rest, t2)

let f e =
  extenv := M.empty;
  let _, t2, _ = g M.empty e in
  (* ǽŪʷ unit ˤʤäƤ뤫 *)
  (try unify Type.Unit t2
   with Unify _ -> failwith "top level does not have type unit");
  extenv := M.map deref_typ !extenv;
  deref_term e
